function [outFile] = combineGPStxtFileParts(inputDir,gpsProd,overwriteOutput)
% This function concatenates gps products into single txt file.

%   [OUTFILE] = combineGPStxtFileParts(INPUTDIR);
%   Reads gps txt files from inputDir with 'SPAN-FINAL' in the name and combines
%   them into single txt OUTFILENAME.Skip if one already exists
%   [OUTFILE] = combineGPStxtFileParts(INPUTDIR,GPSPROD)
%   Reads gps txt files from inputDir with GPSPROD in the name and combines
%   them into single txt OUTFILENAME. Skip if one already exists
%   [OUTFILE] = combineGPStxtFileParts(INPUTDIR,GPSPROD,overwriteOutput)
%   Reads gps txt files from inputDir with GPSPROD in the name and combines
%   them into single txt OUTFILENAME. Overwrites if outputfile if
%   overwriteOutput is true

% It assumes the file names has 'part<#>' in continuous ascending order
% For following files
% RS02_F1005_20161124_SPAN_Final_part1.txt,
% RS02_F1005_20161124_SPAN_Final_part2.txt
% part1 will be read before part2 and so on.

%% Start a log
if strcmpi(get(0,'Diary'),'on')
    logFile = get(0,'DiaryFile');
    localDiary = false;
else
    logFile = 'gpsConvert.log';
    diary(logFile);
    localDiary = true;
end

%% Some housekeeping
if nargin <1
    gpsProd = 'SPAN_FINAL';
    overwriteOutput = false;
elseif nargin < 2
    gpsProd = 'SPAN_FINAL';
    overwriteOutput = false;
elseif nargin < 3
    overwriteOutput = false;
    if isempty(gpsProd)
        gpsProd = 'SPAN_FINAL';
    end
else
    if isempty(gpsProd)
        gpsProd = 'SPAN_FINAL';
    end
end

if ~(overwriteOutput == 0) && ~(overwriteOutput == 1)
    warning('convertGPSStxt2mat:invalidArgType',['Invalid value for endRow (%c) - should be 0(false) or 1(true)\n',...
        'Default value of ''false'' will be used'],overwriteOutput);
    overwriteOutput = false;
end

%% Initialize variables.
delimiter = ' ';
fileFieldSep = '_';
%% Ask for user to select a file if it does not exist or not passed as an argument
if usejava('desktop')
    if ~exist('inputDir','var')
        inputDir= uigetdir('Select the GPS directory to combine files from');
        if isnumeric(inputDir)
            error('convertGPStxt2mat:userAborted','Aborted, no file selected');
        end
    elseif isempty(inputDir) || ~ischar(inputDir) || ~exist(inputDir,'dir')
        inputDir= uigetdir('Select the GPS directory to combine files from');
        if isnumeric(inputDir)
            error('convertGPStxt2mat:userAborted','Aborted, no file selected');
        end
    end
else
    if exist('inputDir','var')
        inputDir = input(['Please enter full path to the GPS file including the extension\n',...
            'e.g.''C:\\Data\\'' or ''/home/data/''\n:']);
        if isnumeric(inputDir) || isempty(inputDir)
            error('convertGPStxt2mat:invalidInput','Aborted! Invalid input');
        end
        if ~(exist(inputDir,'dir')==2)
            error('convertGPStxt2mat:invalidPath','Dir does not exist');
        end
    elseif isempty(inputDir) || ~ischar(inputDir) || ~exist(inputDir,'dir')
        inputDir = input(['Invalid file path\nPlease enter full path to the GPS file including the extension\n',...
            'e.g.''C:\\Data\\'' or ''/home/data/''\n:']);
        if isnumeric(inputDir) || isempty(inputDir)
            error('convertGPStxt2mat:invalidInput','Aborted! Invalid input');
        end
        if ~exist(inputDir,'dir')
            error('convertGPStxt2mat:invalidFilePath','GPS dir does not exist');
        end
    end
end

%% Check existing files and set output file names
gpsFileList = dir([inputDir,filesep,'*',gpsProd,'*.txt']);

if numel(gpsFileList) >1
    fullFileInd = [];
    for gii = 1:numel(gpsFileList)
        [~,gpsFileNameOnly,~] = fileparts(gpsFileList(gii).name);
        gpsFileNameParts = strsplit(gpsFileNameOnly,fileFieldSep);
        if ~any(strncmpi(gpsFileNameParts,'part',4))
            fullFileInd = [fullFileInd gii]; %#ok<AGROW>
            gpsFileList(gii).seq = 0;
        else
            temptxt = gpsFileNameParts{strncmpi(gpsFileNameParts,'part',4)};
            gpsFileList(gii).seq = str2double(temptxt(5:end));
        end
    end
    if ~isempty(fullFileInd)
        if overwriteOutput
            fprintf('convertGPStxt2mat: All existing combined files will be renamed..\n');
            for ddi = 1:numel(fullFileInd)
                movefile([inputDir,filesep,gpsFileList(fullFileInd(ddi)).name],...
                    [inputDir,filesep,gpsFileList(fullFileInd(ddi)).name,datestr(now,'_yyyymmdd_hhMMss')]);
            end
            gpsFileList(fullFileInd) = [];
            
        else
            fprintf('convertGPStxt2mat: A combined txt file already exists...\n');
            if numel(fullFileInd>1)
                warning('convertGPStxt2mat:tooManyOutputFiles','More than one combined files exist');
            end
            outFile = [inputDir,filesep,gpsFileList(fullFileInd(1)).name];
            if localDiary
                diary off;
                delete(logFile);
            end
            return;
        end
    end
    
    [~,sx] = sort([gpsFileList.seq]);
    gpsFileList = gpsFileList(sx);
    
    [~,gpsFileNameOnly,~] = fileparts(gpsFileList(1).name);
    gpsFileNameParts = strsplit(gpsFileNameOnly,fileFieldSep);
    gpsFileNameParts(strncmpi(gpsFileNameParts,'part',4)) = []; % remove component that has 'part#' in it
    outFile = gpsFileNameParts{1};
    for gfn = 2:numel(gpsFileNameParts)
        outFile = [outFile,fileFieldSep,gpsFileNameParts{gfn}]; %#ok<AGROW>
    end
    outFile = [inputDir,filesep,outFile,'.txt'];
else
    fprintf('convertGPStxt2mat: Only one txt file exist...\n');
    if localDiary
        diary off;
        delete(logFile);
    end
    return;
end

logPath = fullfile(inputDir,filesep,'..',filesep,'log',filesep);

if(exist(outFile,'file') && ~overwriteOutput)
    fprintf('convertGPStxt2mat: Combined txt file already exists...\n');
    if localDiary
        diary off;
        delete(logFile);
    end
    return;
else
    fprintf('convertGPStxt2mat: Combining GPS txt files...\n')
end


%% Format string for each line of text: use this to identify data
%   column1: date strings (%s) -mm/dd/yyyy (skip)
%	column2: double (%f) utc time of day sec (skip)
%   column3: time string (%s) - HH:MM:SS (skip)
%	column4: double (%f) - unix time seconds
%   column5: double (%f) - Latitude
%	column6: double (%f) - Longitude
%   column7: double (%f) - Ellipsoid Height H-Ell
%	column8: double (%f) - Pitch
%	column9: double (%f) - Roll
%	column10: double (%f) - Heading
% For more information, see the TEXTSCAN documentation.
formatSpec = '%*s%*f%*s%f%f%f%f%f%f%f%[^\n\r]';

%% copy the first file and open for appending
copyfile([inputDir,filesep,gpsFileList(1).name],outFile);
fprintf('convertGPStxt2mat: %s\n',[inputDir,filesep,gpsFileList(1).name]);
outFId = fopen(outFile,'a');
OutFilecleaner = onCleanup(@()fclose(outFId));
for fjj = 2:numel(gpsFileList)
    fprintf('convertGPStxt2mat: %s\n',[inputDir,filesep,gpsFileList(fjj).name]);
    %% read rest of the files and append to output file
    %% Open the text file.
    inFId = fopen([inputDir,filesep,gpsFileList(fjj).name],'r');
    InFilecleaner = onCleanup(@()fclose(inFId));
    %% Read columns of data according to format string.
    % This call is based on the structure of the file used to generate this
    % code. If an error occurs for a different file, try regenerating the code
    % from the Import Tool.
    
    for nhh = 1:100
        try
            textscan(inFId, formatSpec, 1, 'Delimiter', delimiter, 'MultipleDelimsAsOne', true, 'EmptyValue' ,NaN,'HeaderLines', nhh, 'ReturnOnError', false);
            frewind(inFId);
            break;
        catch
            frewind(inFId);
        end
    end
    
    if nhh<100
        for fll = 1:nhh
            fgetl(inFId);
        end
        data = fread(inFId);
        clear InFilecleaner;
        fwrite(outFId,data);
    else
        warning('convertGPStxt2mat:tooManyOutputFiles','Aborted- End of header not found!');
    end
end
fprintf('convertGPStxt2mat: Created combined file: %s\n',outFile);
clear OutFilecleaner;
disp('convertGPStxt2mat: Done!');
[~,logFileNameOnly,logFileExt] = fileparts(logFile);
if ~exist(logPath,'dir')
    mkdir(logPath);
end
if localDiary
    diary off;
    movefile(logFile,[logPath,filesep,logFileNameOnly,datestr(now,'_yyyymmdd_hhMMss'),logFileExt]);
else
    diary off;
    copyfile(logFile,[logPath,filesep,logFileNameOnly,datestr(now,'_yyyymmdd_hhMMss'),logFileExt]);
    diary on;
end
end