function [outfile] = convertGPStxt2mat(fileName, startRow, endRow,outPath,overwriteOutput)
%   convertGPStext2mat converts processed gps text file to mat file
%   [OUTFILEPATH] = convertGPStxt2mat(INFILENAME);
%   Reads data from text file FILENAME for the default selection.
%
%   [OUTFILEPATH] = convertGPStxt2mat(FILENAME,STARTROW,ENDROW);
%   Reads data from rows STARTROW through ENDROW of text
%   file FILENAME.
%   [OUTFILEPATH] = convertGPStxt2mat(FILENAME,STARTROW,ENDROW,OUTFILEPATH);
%   Reads data from rows STARTROW through ENDROW of text
%   file FILENAME and uses OUTFILEPATH as output file
%
%   Pass an empty matrix [] for any optional argument to use the default
%   settings
%
%   % text file needs to be in following format
%   UTCDate   UTCTime      UTCTime        UTCTime       Latitude      Longitude        H-Ell          Pitch           Roll        Heading
%     (MDY)     (sec)        (HMS)          (sec)          (deg)          (deg)          (m)          (deg)          (deg)          (deg)
%   11/11/2015 73965.300 20:32:45.300 1447273965.300 -77.8680552830 166.9795509225      -40.238   0.8593190000  -3.3753220000 -98.1618430000
% Example:
%   [outfilepath] = convertGPStxt2mat(gpsfilefullpath,25,inf,gpsfileoutputpath);
%
%    See also TEXTSCAN.

% Auto-generated by MATLAB on 2014/08/04 15:05:45
% 02/18/2016 - Modified for New GPS output format (Tej Dhakal)
% 04/29/2016 - Added option to overwrite the output file and ability to
% take empty matrices as optional arguments
%% Start a log
if strcmpi(get(0,'Diary'),'on')
    logFile = get(0,'DiaryFile');
    localDiary = false;
else
    logFile = 'gpsConvert.log';
    diary(logFile);
    localDiary = true;
end

%% Some housekeeping
if nargin <=1
    startRow = 20;
    endRow = inf;
    overwriteOutput = false;
elseif nargin == 2
    if isempty(startRow)
        startRow = 20;
    end
    endRow = inf;
    overwriteOutput = false;
elseif nargin == 3
    if isempty(startRow)
        startRow = 20;
    end
    if isempty(endRow)
        endRow = inf;
    end
    overwriteOutput = false;
elseif nargin == 4
    if isempty(startRow)
        startRow = 20;
    end
    if isempty(endRow)
        endRow = inf;
    end
    if isempty(outPath)
        clear outPath;
    end
    overwriteOutput = false;
elseif nargin == 5
    if isempty(startRow)
        startRow = 20;
    end
    if isempty(endRow)
        endRow = inf;
    end
    if isempty(outPath)
        clear outPath;
    elseif ~ischar(outPath)
        warning('convertGPSStxt2mat:invalidArgType',['Invalid value for outPath (%d) - should be string\n',...
                'Default value of <filepath>/mat will be used'],outPath);
    end
    if isempty(overwriteOutput)
        overwriteOutput = false;
    end
else
    warning('convertGPStxt2mat:invaliNarg','Expected max. 5 arguments. Extra arguments will be ignored')
    if isempty(startRow)
        startRow = 20;
    end
    if isempty(endRow)
        endRow = inf;
    end
    if isempty(outPath)
        clear outPath;
    end
    if isempty(overwriteOutput)
        overwriteOutput = false;
    end
end

% Check argument types and fix if necessary
if ~ isnumeric(startRow)
    warning('convertGPSStxt2mat:invalidArgType',['Invalid value for startRow (%c) - should be non negetive integer\n',...
        'Default value of 20 will be used'],startRow);
    startRow = 20;
elseif startRow < 0
    warning('convertGPSStxt2mat:invalidArgVal',['Invalid value for startRow (%c) - should be non negetive integer\n',...
        'Default value of 20 will be used'],startRow);
    startRow = 20;
end
if ~ isnumeric(endRow)
    warning('convertGPSStxt2mat:invalidArgType',['Invalid value for endRow (%c) - should be non negetive integer or inf\n',...
        'Default value of inf will be used'],endRow);
    endRow = inf;
elseif endRow < 0
    warning('convertGPSStxt2mat:invalidArgVal',['Invalid value for endRow (%c) - should be non negetive integer or inf\n',...
        'Default value of inf will be used'],endRow);
    endRow = inf;
end

if endRow < startRow
    warning('convertGPSStxt2mat:invalidArgVal',['startRow(%d) must be less than endRow(%d)\n',...
        'Default values of 20 & inf will be used'],startRow,endRow);
    startRow = 20;
    endRow = inf;
end

if ~(overwriteOutput == 0) && ~(overwriteOutput == 1)
    warning('convertGPSStxt2mat:invalidArgType',['Invalid value for endRow (%c) - should be 0(false) or 1(true)\n',...
        'Default value of ''false'' will be used'],overwriteOutput);
    overwriteOutput = false;
end

%% Initialize variables.
delimiter = ' ';
%% Ask for user to select a file if it does not exist or not passed as an argument
if usejava('desktop')
    pathname = [];
    if (exist('fileName','var')== 0)
        [fileName,pathname]= uigetfile({'*.txt';'*.csv';'*.xls';'*.xlsx'},'Select the GPS File');
        if isnumeric(fileName) || isnumeric(pathname)
            error('convertGPStxt2mat:userAborted','Aborted, no file selected');
        end
    elseif isempty(fileName) || ~ischar(fileName) || ~(exist(fileName,'file')==2)
            [fileName,pathname]= uigetfile({'*.txt';'*.csv';'*.xls';'*.xlsx'},'Select the GPS File');
            if isnumeric(fileName) || isnumeric(pathname)
                error('convertGPStxt2mat:userAborted','Aborted, no file selected');
            end
    end
    fileName = fullfile(pathname,fileName);
else
    if (exist('fileName','var')== 0)
        fileName = input(['Please enter full path to the GPS file including the extension\n',...
            'e.g.''C:\\Data\\gpsfile.txt'' or ''/home/data/gpsfile.txt''\n:']);
        if isnumeric(fileName) || isempty(fileName)
            error('convertGPStxt2mat:invalidInput','Aborted! Invalid input');
        end
        if ~(exist(fileName,'file')==2)
            error('convertGPStxt2mat:invalidFilePath','File does not exist');
        end
    elseif isempty(fileName) || ~ischar(fileName) || ~(exist(fileName,'file')==2)
        fileName = input(['Invalid file path\nPlease enter full path to the GPS file including the extension\n',...
            'e.g.''C:\\Data\\gpsfile.txt'' or ''/home/data/gpsfile.txt''\n:']);
        if isnumeric(fileName) || isempty(fileName)
            error('convertGPStxt2mat:invalidInput','Aborted! Invalid input');
        end
        if ~(exist(fileName,'file')==2)
            error('convertGPStxt2mat:invalidFilePath','GPS file does not exist');
        end
    end
end
%% Check if the mat file of same name exists. if yes skip processing and return the file name
[pathstr,fname,~] = fileparts(fileName);
if(exist('outPath','var'))
    [outdir,~,outext] = fileparts(outPath);
    if(~strcmpi(outext,'.mat'))
        matfilename = fullfile(outPath,[fname,'.mat']);
    else
        matfilename = outPath;
        outPath = outdir;
    end
    logPath = fullfile(outPath,filesep,'log',filesep);
else
    outPath = [pathstr,filesep,'..',filesep,'mat'];
    matfilename = fullfile(outPath,[fname,'.mat']);
    logPath = fullfile(pathstr,filesep,'..',filesep,'log',filesep);
end

outfile = matfilename;
if(exist(matfilename,'file') && ~overwriteOutput)
    fprintf('convertGPStxt2mat: Using the existing mat file...\n');
    if localDiary
        diary off;
        delete(logFile);
    end
    return;
else
    fprintf('convertGPStxt2mat: Converting GPS text file to mat...\n')
end
% make output directory if it doesn't exist
if ~exist(outPath,'dir')
    mkdir(outPath);
end

%% Format string for each line of text:
%   column1: date strings (%s) -mm/dd/yyyy (skip)
%	column2: double (%f) utc time of day sec (skip)
%   column3: time string (%s) - HH:MM:SS (skip)
%	column4: double (%f) - unix time seconds
%   column5: double (%f) - Latitude
%	column6: double (%f) - Longitude
%   column7: double (%f) - Ellipsoid Height H-Ell
%	column8: double (%f) - Pitch
%	column9: double (%f) - Roll
%	column10: double (%f) - Heading
% For more information, see the TEXTSCAN documentation.
formatSpec = '%*s%*f%*s%f%f%f%f%f%f%f%[^\n\r]';
%% Open the text file.
fileID = fopen(fileName,'r');
cleaner = onCleanup(@()fclose(fileID));
%% Read columns of data according to format string.
% This call is based on the structure of the file used to generate this
% code. If an error occurs for a different file, try regenerating the code
% from the Import Tool.
textscan(fileID, '%[^\n\r]', startRow(1)-1, 'ReturnOnError', false);
dataArray = textscan(fileID, formatSpec, endRow(1)-startRow(1)+1, 'Delimiter', delimiter, 'MultipleDelimsAsOne', true, 'EmptyValue' ,NaN,'HeaderLines', startRow(1)-1, 'ReturnOnError', false);
for block=2:length(startRow)
    frewind(fileID);
    dataArrayBlock = textscan(fileID, formatSpec, endRow(block)-startRow(block)+1, 'Delimiter', delimiter, 'MultipleDelimsAsOne', true, 'EmptyValue' ,NaN,'HeaderLines', startRow(block)-1, 'ReturnOnError', false);
    for col=1:length(dataArray)
        dataArray{col} = [dataArray{col};dataArrayBlock{col}];
    end
end

%% Close the text file.
clear('cleaner');
%% Allocate imported array to column variable names
GPSUTC = dataArray{:, 1};
Lat = dataArray{:, 2};
Lon = dataArray{:, 3};
%% Address a specific issue where waypoint puts an erroneous Longitude during a meridian crossing
% Lon(Lon<0) = Lon(Lon<0)+360;
% dLonInd = find(abs(diff(Lon))>1);    % Any jump more than 1 deg
% dTimeInd = find(abs(diff(GPSUTC)>1));      % Any jump more than 1 s
% dLonInd = dLonInd(~ismember(dLonInd,dTimeInd));     % find all the longitude indices that are not same as time indices
% dLonInd = unique([dLonInd(dLonInd ==1); dLonInd(dLonInd(1:end-1)+1 == dLonInd(2:end));dLonInd(dLonInd ==numel(Lon)-1)]);
% if ~isempty(dLonInd)
%     warning('convertGPStxt2mat:LONJumpsInData','Longitude Jump found in file %s',fileName)
%     for ljj = 1: numel(dLonInd)
%         fprintf('%s\n',datestr(unix2matlabtime(GPSUTC(dLonInd(ljj)))));
%     end
% end
% midInds = dLonInd(dLonInd > 1 & dLonInd<numel(Lon)-1);
% Lon(midInds+1) = 0.5*(Lon(midInds) + Lon(midInds+2));
% % Take care of the case when the jump is the first or last data point
% if any(dLonInd ==1)
%     Lon(1) = interpDegWithJumps(GPSUTC(2:3),Lon(2:3),GPSUTC(1),'linear','extrap');
% end
% if any(dLonInd ==dLonInd<numel(Lon))
%     Lon(end) = interpDegWithJumps(GPSUTC(end-2:end-1),Lon(end-2:end-1),GPSUTC(end),'linear','extrap');
% end
% Lon(Lon>180) = Lon(Lon>180)-360;
Lon = fixLonJumpError(Lon,GPSUTC);
%%
Hell = dataArray{:, 4};
Pitch = dataArray{:,5};
Roll = dataArray{:, 6};
Heading = dataArray{:,7};
unitsNav = struct('GPSUTC','unixsec','Lat','deg','Lon','deg','Hell','m','Pitch','deg','Roll','deg','Heading','deg');
save(matfilename,'GPSUTC','Lat','Lon','Hell','Heading','Pitch','Roll','unitsNav')
fprintf('convertGPStxt2mat: File saved as: %s\n',matfilename);
disp('convertGPStxt2mat: Done!');
[~,logFileNameOnly,logFileExt] = fileparts(logFile);
if ~exist(logPath,'dir')
    mkdir(logPath);
end
if localDiary
    diary off;
    movefile(logFile,[logPath,filesep,logFileNameOnly,datestr(now,'_yyyymmdd_hhMMss'),logFileExt]);
else
    diary off;
    copyfile(logFile,[logPath,filesep,logFileNameOnly,datestr(now,'_yyyymmdd_hhMMss'),logFileExt]);
    diary on;
end
end