function [outfile] = convertLineCsv2mat(fileName, startRow, endRow, outPath, overwriteOutput)
%	Imports Line information data into a mat file from either a comma
%	delimited '.csv' file or a tab delimited '.txt' file
%    [OUTFILEPATH] = convertLineCsv2mat(FILENAME) Reads data from text file FILENAME for the
%   default selection and saves the mat file int he same path as fileName
%    [OUTFILEPATH] = convertLineCsv2mat(FILENAME, STARTROW, ENDROW) Reads data from rows STARTROW
%   through ENDROW of text file FILENAME and saves the mat file in the same
%   path as fileName.
%    [OUTFILEPATH] = convertLineCsv2mat(FILENAME, STARTROW, ENDROW,OUTFILEPATH) Reads data from rows STARTROW
%   through ENDROW of text file FILENAME and saves the mat file in the
%   OUTFILEPATH directory
%   text file columns:
%       FlightNum,LineNum,StartUTCDate,StartUTCTIme,StartUnixTime,StopUTCDate,StopUTCTIme,StopUnixTime,StartLatitude,StartLongitude,StartH-El,StopLatitude,StopLongitude,StopH-El
%       3,L650,11/11/2015,22:11:23,1447279883,11/12/2015,00:57:33,1447289853,838.965,-79.398594,160.547042,1160.628,-78.871971,-154.031139
%
% Example:
%   lineMatpath = convertLinecsv2mat('LineSegsRS1.csv',1, 49,'C:\Test');
%
%    See also TEXTSCAN.

% Auto-generated by MATLAB on 2016/02/19 10:22:24
% 02/19/2016 - Modified (Tej Dhakal)
%% Start a log
if strcmpi(get(0,'Diary'),'on')
    logFile = get(0,'DiaryFile');
    localDiary = false;
else
    logFile = 'lineConvert.log';
    diary(logFile);
    localDiary = true;
end

%% Some housekeeping
if nargin <=1
    startRow = 5;
    endRow = inf;
    overwriteOutput = false;
elseif nargin == 2
    if isempty(startRow)
        startRow = 5;
    end
    endRow = inf;
    overwriteOutput = false;
elseif nargin == 3
    if isempty(startRow)
        startRow = 5;
    end
    if isempty(endRow)
        endRow = inf;
    end
    overwriteOutput = false;
elseif nargin == 4
    if isempty(startRow)
        startRow = 5;
    end
    if isempty(endRow)
        endRow = inf;
    end
    if isempty(outPath)
        clear outPath;
    end
    overwriteOutput = false;
elseif nargin == 5
    if isempty(startRow)
        startRow = 5;
    end
    if isempty(endRow)
        endRow = inf;
    end
    if isempty(outPath)
        clear outPath;
    elseif ~ischar(outPath)
        warning('convertLineCsv2mat:invalidArgType',['Invalid value for outPath (%d) - should be string\n',...
                'Default value of <filepath>/mat will be used'],outPath);
    end
    if isempty(overwriteOutput)
        overwriteOutput = false;
    end
else
    warning('convertLineCsv2mat:invalidNarg','Expected max. 5 arguments. Extra arguments will be ignored')
    if isempty(startRow)
        startRow = 5;
    end
    if isempty(endRow)
        endRow = inf;
    end
    if isempty(outPath)
        clear outPath;
    end
    if isempty(overwriteOutput)
        overwriteOutput = false;
    end
end

% Check argument types and fix if necessary
if ~ isnumeric(startRow)
    warning('convertLineCsv2mat:invalidArgType',['Invalid value for startRow (%c) - should be non negetive integer\n',...
        'Default value of 5 will be used'],startRow);
    startRow = 5;
elseif startRow < 0
    warning('convertLineCsv2mat:invalidArgVal',['Invalid value for startRow (%c) - should be non negetive integer\n',...
        'Default value of 5 will be used'],startRow);
    startRow = 5;
end
if ~ isnumeric(endRow)
    warning('convertLineCsv2mat:invalidArgType',['Invalid value for endRow (%c) - should be non negetive integer or inf\n',...
        'Default value of inf will be used'],endRow);
    endRow = inf;
elseif endRow < 0
    warning('convertLineCsv2mat:invalidArgVal',['Invalid value for endRow (%c) - should be non negetive integer or inf\n',...
        'Default value of inf will be used'],endRow);
    endRow = inf;
end

if endRow < startRow
    warning('convertLineCsv2mat:invalidArgVal',['startRow(%d) must be less than endRow(%d)\n',...
        'Default values of 5 & inf will be used'],startRow,endRow);
    startRow = 5;
    endRow = inf;
end

if ~(overwriteOutput == 0) && ~(overwriteOutput == 1)
    warning('convertLineCsv2mat:invalidArgType',['Invalid value for endRow (%c) - should be 0(false) or 1(true)\n',...
        'Default value of ''false'' will be used'],overwriteOutput);
    overwriteOutput = false;
end

%% Ask for user to select a file if it does not exist or not passed as an argument
if usejava('desktop')
    pathname = [];
    if (exist('fileName','var')== 0)
        [fileName,pathname]= uigetfile({'*.txt';'*.csv'},'Select the Line File');
        if isnumeric(fileName) || isnumeric(pathname)
            error('convertLineCsv2mat:userAborted','Aborted, no file selected');
        end
    elseif isempty(fileName) || ~ischar(fileName) || ~(exist(fileName,'file')==2)
            [fileName,pathname]= uigetfile({'*.txt';'*.csv'},'Select the Line File');
            if isnumeric(fileName) || isnumeric(pathname)
                error('convertLineCsv2mat:userAborted','Aborted, no file selected');
            end
    end
    fileName = fullfile(pathname,fileName);
else
    if (exist('fileName','var')== 0)
        fileName = input(['Please enter full path to the Line file including the extension\n',...
            'e.g.''C:\\Data\\Linefile.txt'' or ''/home/data/Linefile.txt''\n:']);
        if isnumeric(fileName) || isempty(fileName)
            error('convertLineCsv2mat:invalidInput','Aborted! Invalid input');
        end
        if ~(exist(fileName,'file')==2)
            error('convertLineCsv2mat:invalidFilePath','File does not exist');
        end
    elseif isempty(fileName) || ~ischar(fileName) || ~(exist(fileName,'file')==2)
        fileName = input(['Invalid file path\nPlease enter full path to the Line file including the extension\n',...
            'e.g.''C:\\Data\\Linefile.txt'' or ''/home/data/Linefile.txt''\n:']);
        if isnumeric(fileName) || isempty(fileName)
            error('convertLineCsv2mat:invalidInput','Aborted! Invalid input');
        end
        if ~(exist(fileName,'file')==2)
            error('convertLineCsv2mat:invalidFilePath','Line file does not exist');
        end
    end
end

%% Check if the mat file of same name exists. if yes skip processing and return the file name
[pathstr,fname,fext] = fileparts(fileName);
if(exist('outPath','var'))
    matfileName = fullfile(outPath,[fname,'.mat']);
    logPath = fullfile(outPath,filesep,'..',filesep,'log',filesep);
else
    outPath = [pathstr,filesep,'..',filesep,'mat'];
    matfileName = fullfile(outPath,filesep,[fname,'.mat']);
    logPath = fullfile(pathstr,filesep,'..',filesep,'log',filesep);
end

outfile = matfileName;
if(exist(matfileName,'file') && ~overwriteOutput)
    fprintf('convertLinesCsv2mat: Using the existing mat file...\n');
    if localDiary
        diary off;
        delete(logFile);
    end
    return;
else
    fprintf('convertLinesCsv2mat: Converting Line text file to mat...\n')
end
% make output directory if it doesn't exist
if ~exist(outPath,'dir')
    mkdir(outPath);
end

%% Read columns of data as strings:
formatSpec = '%f%s%*s%*s%f%*s%*s%f%f%f%f%f%f%f%s%f%[^\n\r]';
%% Format string for each line of text:
%   column1: double (%f)
%	column2: text (%s)
%	column2: text (%s) -skip
%	column2: text (%s) -skip
%   column5: double (%f)
%	column2: text (%s) -skip
%	column2: text (%s) -skip
%	column8: double (%f)
%   column9: double (%f)
%	column10: double (%f)
%   column11: double (%f)
%	column12: double (%f)
%   column13: double (%f)
%	column14: double (%f)
%   column15: text   (%s)
% For more information, see the TEXTSCAN documentation.

%% Open the text file.
if (strcmp(fext,'.txt'))
    delimiter = '\t';
elseif(strcmp(fext,'.csv'))
    delimiter = ',';
else
    error('convertLineCsv2mat:InvalidInputFileType','Invalid file type, select a csv or a txt line File')
end

fileID = fopen(fileName,'r');
cleaner = onCleanup(@()fclose(fileID));
%% Read columns of data according to format string.
% This call is based on the structure of the file used to generate this
% code. If an error occurs for a different file, try regenerating the code
% from the Import Tool.
dataArray = textscan(fileID, formatSpec, endRow(1)-startRow(1)+1, 'Delimiter', delimiter, 'HeaderLines', startRow(1)-1, 'ReturnOnError', false);
for block=2:length(startRow)
    frewind(fileID);
    dataArrayBlock = textscan(fileID, formatSpec, endRow(block)-startRow(block)+1, 'Delimiter', delimiter, 'HeaderLines', startRow(block)-1, 'ReturnOnError', false);
    for col=1:length(dataArray)
        dataArray{col} = [dataArray{col};dataArrayBlock{col}];
    end
end

%% Close the text file.
clear('cleaner');
%% Post processing for unimportable data.
% No unimportable data rules were applied during the import, so no post
% processing code is included. To generate code which works for
% unimportable data, select unimportable cells in a file and regenerate the
% script.

%% Allocate imported array to column variable names
FlightNum = dataArray{:, 1};
LineNum = dataArray{:, 2};
LineTime = [dataArray{:, 3} dataArray{:, 4}];   % [starttime stoptime]
LineLat = [dataArray{:, 5} dataArray{:, 8}];
LineLon = [dataArray{:, 6} dataArray{:, 9}];
LineHell = [dataArray{:, 7} dataArray{:, 10}];
LineProj = dataArray{:,11};
LineStartElev = dataArray{:,12};
save(matfileName,'FlightNum','LineNum','LineTime','LineLat','LineLon','LineHell','LineProj','LineStartElev')
fprintf('convertLinesCsv2mat: File saved as: %s\n',matfileName);
disp('convertGPStxt2mat: Done!');
[~,logFileNameOnly,logFileExt] = fileparts(logFile);
if ~exist(logPath,'dir')
    mkdir(logPath);
end
if localDiary
    diary off;
    movefile(logFile,[logPath,filesep,logFileNameOnly,datestr(now,'_yyyymmdd_hhMMss'),logFileExt]);
else
    diary off;
    copyfile(logFile,[logPath,filesep,logFileNameOnly,datestr(now,'_yyyymmdd_hhMMss'),logFileExt]);
    diary on;
end
    
end
