function [pksOut,locsOut] = findpeaksMat(Din,nPeaks,pickOption,thresPeak)
%FINDPEAKS Find local peaks in data
%   PKS = FINDPEAKS(X) finds local peaks in the data vector X. if X is a
%   Matrix, FINDPEAKS(X) returns local peaks in each column of X. If the
%   number of peaks for each columns are different, In the PKS matrix, the 
%   the columns with fewer number of peaks are padded with NaNs
%   A local peak is defined as a data sample which is either larger than
%   the two neighboring samples or is equal to Inf.
%   [PKS,LOCS]= FINDPEAKS(X) also returns the indices LOCS at which the
%   peaks occur.If X is a matrix and If the number of peaks for each column
%   are are different, In the LOCS matrix, the the columns with fewer
%   number  of peaks are padded with NaNs
%
%   [...] = FINDPEAKS(X,NPEAKS) specifies the maximum number of
%   peaks to be found. NP is an integer greater than zero. If not
%   specified or [], all peaks are returned.
%
%   [...] = FINDPEAKS(X,NPEAKS,POPT) specifies the direction of sorting
%   of peaks. STR can take values of 'ascend','descend' or 'first'. If not
%   specified, STR takes the value of 'first' and the peaks are returned in
%   the order of their occurrence. This option is applied before returning
%   NPEAKS
%
%   [...] = FINDPEAKS(X,NPEAKS,POPT,MPH) finds only those peaks that
%   are greater than the minimum peak height, MPH. Specifying a minimum
%   peak height may help in reducing the processing time. MPH is a real
%   valued scalar. The default value of MPH is -Inf. This option is applied
%   before NPEAKS and POPT



if isrow(Din)
    Din = Din(:);
end
if all(isnan(Din))
    warning('findpeaksMat:allNanData','Input Data are all NaNs')
    pksOut = ones(nPeaks,size(Din,2))*NaN;
    locsOut = ones(nPeaks,size(Din,2))*NaN;
    return;
end
pksOut = ones(size(Din))*NaN;
locsOut = ones(size(Din))*NaN;
numPeaks = zeros(1,size(Din,2));
if ~exist('thresPeak','var')
    thresPeak = -Inf;
elseif isempty(thresPeak)
    thresPeak = -Inf;
end


if ~ exist('pickOption','var')
    pickOption = 'first';
elseif isempty(pickOption)
    pickOption = 'first';
end

for ii = 1:size(Din,2)
    thresInds = find(Din(:,ii)>thresPeak);
    
    if(isempty(thresInds))
        warning('findpeaksMat:largeMinPeakHeight', 'Threshold too high - no peaks found');
        warning('off','findpeaksMat:largeMinPeakHeight');
        continue;
    end
    
    trend = sign(diff(Din(:,ii)));
    N = size(trend,1);
    idx = find(trend==0); % Find flats
    idxm1 = find(trend(min(idx+1,N)) <0);
    while(~isempty(idxm1))
        trend(idx(idxm1)) = -1;
        idx = find(trend==0); % Find flats
        idxm1 = find(trend(min(idx+1,N)) <0);
    end
    trend(idx(trend(min(idx+1,N))>=0)) = 1;
    idxp = find(diff(trend)==-2)+1;  % Get all the peaks
    if ~isempty(idxp)
        locstmp = intersect(thresInds,idxp(:));      % Keep peaks above MinPeakHeight
        pkstmp  = Din(locstmp,ii);
    else
        warning('findpeaksMat:noPeaks','No peaks found for column %d',ii)
        warning('off','findpeaksMat:noPeaks');
        continue;
    end
    
    if ~ strcmp(pickOption,'first')
        [pkstmp,tempInd] = sort(pkstmp,pickOption);
        locstmp = locstmp(tempInd);
    end
    numPeaks(1,ii) = size(pkstmp,1);
    pksOut(1:numPeaks(1,ii),ii) = pkstmp;
    locsOut(1:numPeaks(1,ii),ii) = locstmp;
end
if exist('nPeaks','var')
    if ~isempty(nPeaks)
        pksOut = pksOut(1:nPeaks,:);
        locsOut = locsOut(1:nPeaks,:);
        if any(numPeaks < nPeaks)
            warning('findpeaksMat:notEnoughPeaks','Number of peaks available are less than %d for some or all',nPeaks)
            warning('off','findpeaksMat:notEnoughPeaks');
        end
    else
        pksOut = pksOut(1:max(numPeaks),:);
        locsOut = locsOut(1:max(numPeaks),:);
    end
else
    pksOut = pksOut(1:max(numPeaks),:);
    locsOut = locsOut(1:max(numPeaks),:);
end
warning('on','findpeaksMat:notEnoughPeaks');
warning('on','findpeaksMat:noPeaks');
warning('on','findpeaksMat:largeMinPeakHeight');
end