function DataOut=intuitive_lowpass2(DataIn,minxwavelength,minywavelength)

% Intuitive Lowpass Filter, 2D
% By Mike Wolovick, 1/1/2011 (recreation of a script began in April).

% This function low-pass filters an input with a gaussian filter.

% The standard deviation of the gaussian is one-half minwavelength (so
% minwavelength is roughly full width at half max).

% minwavelength must be expressed in terms of number of samples.  It must
% be greater than one; however, it is allowed to have non-integer values.

% DataIn is padded with a buffer determined by a best-fit plane in order to
% minimize edge effects.

% A linear best-fit to DataIn is removed before the filtering is performed,
% then added back to DataOut at the end.

% This function assumes that the input DataIn is evenly spaced

% Input checking:
if size(DataIn,1)<2 || size(DataIn,2)<2
    error('Input "DataIn" must be at least a 2x2 matrix.')
end
if isscalar(minxwavelength)==0
    error('Input "minxwavelength" must be a scalar')
elseif isreal(minxwavelength)==0
    error('Input "minxwavelength" must be real-valued.')
elseif minxwavelength<1
    error('Input "minxwavelength" must be greater than one.')
end
if isscalar(minywavelength)==0
    error('Input "minywavelength" must be a scalar')
elseif isreal(minywavelength)==0
    error('Input "minywavelength" must be real-valued.')
elseif minywavelength<1
    error('Input "minywavelength" must be greater than one.')
end

% Define Parameters:
lx=size(DataIn,2);
ly=size(DataIn,1);
ltot=lx*ly;
integerxwavelength=round(minxwavelength);
integerywavelength=round(minywavelength);
lxbig=lx+6*integerxwavelength;
lybig=ly+6*integerywavelength;
X=linspace(1,lx,lx);
Y=linspace(1,ly,ly)';
[X_grid,Y_grid]=meshgrid(X,Y);
Xbig=linspace(1,lxbig,lxbig);
Ybig=linspace(1,lybig,lybig)';
[Xbig_grid,Ybig_grid]=meshgrid(Xbig,Ybig);
xcent=mean(Xbig);
ycent=mean(Ybig);
stdevx=.5*minxwavelength;
stdevy=.5*minywavelength;
% Remove A Best-Fit Plane:
A=[ones(ltot,1),reshape(X_grid,ltot,1),reshape(Y_grid,ltot,1)];
squareA=A'*A;
equation=squareA\A'*reshape(DataIn,ltot,1);
BestPlane=equation(1)+equation(2)*X_grid+equation(3)*Y_grid;
DataIn=DataIn-BestPlane;
clear A squareA equation
% Perform Filtering:
DataSpread=zeros(lybig,lxbig);
DataSpread(3*integerywavelength+1:end-3*integerywavelength,3*integerxwavelength+1:end-3*integerxwavelength)=DataIn;
[DataSpread(1:3*integerywavelength,3*integerxwavelength+1:end-3*integerxwavelength),blank1]=meshgrid(DataIn(1,:),zeros(3*integerywavelength,1));
[DataSpread(end-3*integerywavelength+1:end,3*integerxwavelength+1:end-3*integerxwavelength),blank2]=meshgrid(DataIn(end,:),zeros(3*integerywavelength,1));
[blank3,DataSpread(3*integerywavelength+1:end-3*integerywavelength,1:3*integerxwavelength)]=meshgrid(zeros(1,3*integerxwavelength),DataIn(:,1));
[blank4,DataSpread(3*integerywavelength+1:end-3*integerywavelength,end-3*integerxwavelength+1:end)]=meshgrid(zeros(1,3*integerxwavelength),DataIn(:,end));
DataSpread(1:3*integerywavelength,1:3*integerxwavelength)=DataIn(1,1);
DataSpread(end-3*integerywavelength+1:end,1:3*integerxwavelength)=DataIn(end,1);
DataSpread(1:3*integerywavelength,end-3*integerxwavelength+1:end)=DataIn(1,end);
DataSpread(end-3*integerywavelength+1:end,end-3*integerxwavelength+1:end)=DataIn(end,end);
clear blank*
Filter=exp(-.5*((Xbig_grid-xcent)/stdevx).^2).*exp(-.5*((Ybig_grid-ycent)/stdevy).^2);
Filter=Filter/sum(sum(Filter));
DataOut=fftshift(ifft2(fft2(DataSpread).*fft2(Filter)));
DataOut=DataOut(3*integerywavelength:end-3*integerywavelength-1,3*integerxwavelength:end-3*integerxwavelength-1)+BestPlane;
