function DataOut=intuitive_lowpass3(DataIn,minxwavelength,minywavelength,minzwavelength)

% Intuitive Lowpass Filter, 3D
% By Mike Wolovick, 4/23/2011

% This function low-pass filters an input with a gaussian filter.

% The standard deviation of the gaussian is one-half minwavelength (so
% minwavelength is roughly full width at half max).

% The indexing of DataIn goes by (y,x,z) 

% minwavelength must be expressed in terms of number of samples.  It must
% be greater than one; however, it is allowed to have non-integer values.

% DataIn is padded with a buffer determined the estension of the endpoints
% along a linear best-fit in order to minimize edge effects.

% This function assumes that the input DataIn is evenly spaced

% Input checking:
if ndims(DataIn)~=3
    error('Input "DataIn" must be a 3-dimensional matrix.')
end
if isscalar(minxwavelength)==0
    error('Input "minxwavelength" must be a scalar')
elseif isreal(minxwavelength)==0
    error('Input "minxwavelength" must be real-valued.')
elseif minxwavelength<1
    error('Input "minxwavelength" must be greater than one.')
end
if isscalar(minywavelength)==0
    error('Input "minywavelength" must be a scalar')
elseif isreal(minywavelength)==0
    error('Input "minywavelength" must be real-valued.')
elseif minywavelength<1
    error('Input "minywavelength" must be greater than one.')
end
if isscalar(minzwavelength)==0
    error('Input "minzwavelength" must be a scalar')
elseif isreal(minzwavelength)==0
    error('Input "minzwavelength" must be real-valued.')
elseif minzwavelength<1
    error('Input "minzwavelength" must be greater than one.')
end

% Define Parameters:
lx=size(DataIn,2);
ly=size(DataIn,1);
lz=size(DataIn,3);
ltot=lx*ly*lz;
integerxwavelength=round(minxwavelength);
integerywavelength=round(minywavelength);
integerzwavelength=round(minzwavelength);
lxbig=lx+6*integerxwavelength;
lybig=ly+6*integerywavelength;
lzbig=lz+6*integerzwavelength;

% Define Coordinates:
X=linspace(1,lx,lx);
Y=linspace(1,ly,ly)';
Z=linspace(1,lz,lz);
[X_grid,Y_grid,Z_grid]=meshgrid(X,Y,Z);
Xbig=linspace(1,lxbig,lxbig);
Ybig=linspace(1,lybig,lybig)';
Zbig=linspace(1,lzbig,lzbig);
[Xbig_grid,Ybig_grid,Zbig_grid]=meshgrid(Xbig,Ybig,Zbig);
xcent=mean(Xbig);
ycent=mean(Ybig);
zcent=mean(Zbig);
stdevx=.5*minxwavelength;
stdevy=.5*minywavelength;
stdevz=.5*minzwavelength;

% Remove A Best-Fit Plane:
A=[ones(ltot,1),reshape(X_grid,ltot,1),reshape(Y_grid,ltot,1),reshape(Z_grid,ltot,1)];
squareA=A'*A;
equation=squareA\A'*reshape(DataIn,ltot,1);
BestLinear=equation(1)+equation(2)*X_grid+equation(3)*Y_grid+equation(4)*Z_grid;
DataIn=DataIn-BestLinear;
clear A squareA equation

% Add Buffers to Data:
DataSpread=zeros(lybig,lxbig,lzbig);
DataSpread(3*integerywavelength+1:end-3*integerywavelength,3*integerxwavelength+1:end-3*integerxwavelength,3*integerzwavelength+1:end-3*integerzwavelength)=DataIn;  % central data portion
% Side Buffers:
for d1=1:3*integerywavelength
    DataSpread(d1,3*integerxwavelength+1:end-3*integerxwavelength,3*integerzwavelength+1:end-3*integerzwavelength)=DataIn(1,:,:);
end
for d1=lybig-3*integerywavelength+1:lybig
    DataSpread(d1,3*integerxwavelength+1:end-3*integerxwavelength,3*integerzwavelength+1:end-3*integerzwavelength)=DataIn(end,:,:);
end
for d2=1:3*integerxwavelength
    DataSpread(3*integerywavelength+1:end-3*integerywavelength,d2,3*integerzwavelength+1:end-3*integerzwavelength)=DataIn(:,1,:);
end
for d2=lxbig-3*integerxwavelength+1:lxbig
    DataSpread(3*integerywavelength+1:end-3*integerywavelength,d2,3*integerzwavelength+1:end-3*integerzwavelength)=DataIn(:,end,:);
end
for d3=1:3*integerzwavelength
    DataSpread(3*integerywavelength+1:end-3*integerywavelength,3*integerxwavelength+1:end-3*integerxwavelength,d3)=DataIn(:,:,1);
end
for d3=lzbig-3*integerzwavelength+1:lzbig
    DataSpread(3*integerywavelength+1:end-3*integerywavelength,3*integerxwavelength+1:end-3*integerxwavelength,d3)=DataIn(:,:,end);
end
% Edge Buffers:
[blank1,blank2,DataSpread(1:3*integerywavelength,1:3*integerxwavelength,3*integerzwavelength+1:end-3*integerzwavelength)]=ndgrid(zeros(1,3*integerywavelength),zeros(1,3*integerxwavelength)',DataIn(1,1,:));  % x-,y-
[blank1,blank2,DataSpread(end-3*integerywavelength+1:end,1:3*integerxwavelength,3*integerzwavelength+1:end-3*integerzwavelength)]=ndgrid(zeros(1,3*integerywavelength),zeros(1,3*integerxwavelength)',DataIn(end,1,:));  % x-,y+
[DataSpread(3*integerywavelength+1:end-3*integerywavelength,1:3*integerxwavelength,1:3*integerzwavelength),blank1,blank2]=ndgrid(DataIn(:,1,1),zeros(1,3*integerxwavelength)',zeros(1,3*integerzwavelength));  % x-,z-
[DataSpread(3*integerywavelength+1:end-3*integerywavelength,1:3*integerxwavelength,end-3*integerzwavelength+1:end),blank1,blank2]=ndgrid(DataIn(:,1,end),zeros(1,3*integerxwavelength)',zeros(1,3*integerzwavelength));  % x-,z+
[blank1,blank2,DataSpread(1:3*integerywavelength,end-3*integerxwavelength+1:end,3*integerzwavelength+1:end-3*integerzwavelength)]=ndgrid(zeros(1,3*integerywavelength),zeros(1,3*integerxwavelength)',DataIn(1,end,:));  % x+,y-
[blank1,blank2,DataSpread(end-3*integerywavelength+1:end,end-3*integerxwavelength+1:end,3*integerzwavelength+1:end-3*integerzwavelength)]=ndgrid(zeros(1,3*integerywavelength),zeros(1,3*integerxwavelength)',DataIn(end,end,:));  % x+,y+
[DataSpread(3*integerywavelength+1:end-3*integerywavelength,end-3*integerxwavelength+1:end,1:3*integerzwavelength),blank1,blank2]=ndgrid(DataIn(:,end,1),zeros(1,3*integerxwavelength)',zeros(1,3*integerzwavelength));  % x+,z-
[DataSpread(3*integerywavelength+1:end-3*integerywavelength,end-3*integerxwavelength+1:end,end-3*integerzwavelength+1:end),blank1,blank2]=ndgrid(DataIn(:,end,end),zeros(1,3*integerxwavelength)',zeros(1,3*integerzwavelength));  % x+,z+
[blank1,DataSpread(1:3*integerywavelength,3*integerxwavelength+1:end-3*integerxwavelength,1:3*integerzwavelength),blank2]=ndgrid(zeros(1,3*integerywavelength),DataIn(1,:,1),zeros(1,3*integerzwavelength));  % y-,z-
[blank1,DataSpread(1:3*integerywavelength,3*integerxwavelength+1:end-3*integerxwavelength,end-3*integerzwavelength+1:end),blank2]=ndgrid(zeros(1,3*integerywavelength),DataIn(1,:,end),zeros(1,3*integerzwavelength));  % y-,z+
[blank1,DataSpread(end-3*integerywavelength+1:end,3*integerxwavelength+1:end-3*integerxwavelength,1:3*integerzwavelength),blank2]=ndgrid(zeros(1,3*integerywavelength),DataIn(end,:,1),zeros(1,3*integerzwavelength));  % y+,z-
[blank1,DataSpread(end-3*integerywavelength+1:end,3*integerxwavelength+1:end-3*integerxwavelength,end-3*integerzwavelength+1:end),blank2]=ndgrid(zeros(1,3*integerywavelength),DataIn(end,:,end),zeros(1,3*integerzwavelength));  % y+,z+
% Corner Buffers:
DataSpread(1:3*integerywavelength,1:3*integerxwavelength,1:3*integerzwavelength)=DataIn(1,1,1);  % y-,x-,z-
DataSpread(end-3*integerywavelength+1:end,1:3*integerxwavelength,1:3*integerzwavelength)=DataIn(end,1,1);  % y+,x-,z-
DataSpread(1:3*integerywavelength,end-3*integerxwavelength+1:end,1:3*integerzwavelength)=DataIn(1,end,1);  % y-,x+,z-
DataSpread(end-3*integerywavelength+1:end,end-3*integerxwavelength+1:end,1:3*integerzwavelength)=DataIn(end,end,1);  % y+,x+,z-
DataSpread(1:3*integerywavelength,1:3*integerxwavelength,end-3*integerzwavelength+1:end)=DataIn(1,1,end);  % y-,x-,z+
DataSpread(end-3*integerywavelength+1:end,1:3*integerxwavelength,end-3*integerzwavelength+1:end)=DataIn(end,1,end);  % y+,x-,z+
DataSpread(1:3*integerywavelength,end-3*integerxwavelength+1:end,end-3*integerzwavelength+1:end)=DataIn(1,end,end);  % y-,x+,z+
DataSpread(end-3*integerywavelength+1:end,end-3*integerxwavelength+1:end,end-3*integerzwavelength+1:end)=DataIn(end,end,end);  % y+,x+,z+
clear blank*

% Perform Filtering:
Filter=exp(-.5*((Xbig_grid-xcent)/stdevx).^2).*exp(-.5*((Ybig_grid-ycent)/stdevy).^2).*exp(-.5*((Zbig_grid-zcent)/stdevz).^2);
Filter=Filter/sum(sum(sum(Filter)));
DataOut=fftshift(ifftn(fftn(DataSpread).*fftn(Filter)));
DataOut=DataOut(3*integerywavelength:end-3*integerywavelength-1,3*integerxwavelength:end-3*integerxwavelength-1,3*integerzwavelength:end-3*integerzwavelength-1)+BestLinear;
