function [] = lasSurfElev(inputDir,gpsFile,outputDir,gridOpt,savePointcloud,plotPointcloud,overwriteOutput,filterWindow,thresholdDeviation,maxElevChangeWin,maxElevChangeLine,maxGeoid,minGeoid)
% This extracts the ellipsoidal height of ground track for the flight from a las file
% inputDir: source directory for las files to process.
% gpsFile : full path to gps mat file for the flight, user will be
%           prompted if not specified.
% outputDir (optional): destination directory for processed mat file. If
%           not specified or empty ([])  a 'mat' directory is created in the inputDir
% GRIDOPT (optional): 'nearest' (default),'linear','natural' or 'cubic' see
%                       help griddata
% savePointcloud (optional): if true, saves a matfile with all data from
%                las files. false by default
% plotPointcloud (optional); if true, plots a decimated point cloud of all
%                           records and saves a png. true by default
% OVERWRITEOUTPUT (optional); if true, overwrites any existing output mat
%                 files. true by default
% FILTERWINDOW (optional): s, Time window without a drastic change in
%               surface elevation, default  = 1s
% THRESHOLDDEVIATION (optional): m, Threshold Standard deviaiton of surface
%                     elevation after which the filter kicks in. default = 5 m.
% MAXELEVCHANGEWIN (optional): m, Maximum expected elevation change +/- in
%                   the FILTERWINDOW, default = 20 m.
% MAXELEVCHANGELINE (optional): m, Maximum expected elevation change +/- in
%                   the entire Line, default: 250 m
% MAXGEOID (optional): m, Maximum expected geoid height (max(geoid -
%           ellipsoid)). Any points ouside this will be filtered out default +inf
% MINGEOID (optional): m, Minimum expected geoid height (min(geoid -
% ellipsoid). Any points ouside this will be filtered out. default -inf;
% dependencies: lasdata.m, adjStdGPStoUTC.m,getGPSLeapSecs.m,tai-utc.dat

% Notes:
%   For particularly bad data, use maxGeoid,minGeoid and
%       maxElevChangeWin to get better results.
%   Use [] for unused/default parameter. e.g.
%       lasSurfElev('C/inputDir/','c/gpsFile.mat',[],'linear',false,true,[],[],[],300,[],-200)

% This function makes some assumptions about the las file.
% Use these settings to export data on RiProcess:
%     Format v1.4 without waveform data
%     Source Coordinate System (WGS84)
%     Geographic (Lon,Lat,Alt)
%     Export Original Points
%     Adjusted GPS Time (GPS Time -1e9 sec.)
%     Angle Filter +/-5 deg or narrower (85 to 95) for faster processing

% Outputs:
%   Ground track file with: Lat, Lon, Hell, PointReflectance SurfHell at the same rate as gps input file
%   PointReflectance is a 16 nit integer - what it exactly represents needs
%   to be investigated.
%   Optionallly, point cloud mat files with :
%   pointHell: m, Laser Elevation (Ellipsoid WGS84) original points from las file.
%   pointLon : deg, Longitude from las file
%   pointLat: det, Latitude for points from las file
%   pointLat: utc time from las file (seconds appears to corrected for leap seconds already)
%   pointAngle: deg, Scan angle for the point
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% some housekeeping
if ~exist('inputDir','var')
    inputDir = [];
end
if ~exist('outputDir','var')
    outputDir = [];
end
if ~exist('gpsFile','var')
    gpsFile = [];
end
if ~exist('savePointcloud','var')
    savePointcloud = [];
end
if ~exist('plotPointcloud','var')
    plotPointcloud = [];
end
if ~exist('overwriteOutput','var')
    overwriteOutput = [];
end
if ~exist('filterWindow','var')
    filterWindow = [];
end
if ~exist('thresholdDeviation','var')
    thresholdDeviation = [];
end
if ~exist('maxElevChangeWin','var')
    maxElevChangeWin = [];
end
if ~exist('maxElevChangeLine','var')
    maxElevChangeLine = [];
end
if ~exist('maxGeoid','var')
    maxGeoid = [];
end
if ~exist('minGeoid','var')
    minGeoid = [];
end
if ~exist('gridOpt','var')
    gridOpt = [];
end

% assign default vaues to empty variables
if isempty(savePointcloud)
    savePointcloud = false;
end
if isempty(plotPointcloud)
    plotPointcloud = true;
end
if isempty(overwriteOutput)
    overwriteOutput = true;
end
if isempty(filterWindow)
    filterWindow = 1;
end
if isempty(thresholdDeviation)
    thresholdDeviation = 5;
end
if isempty(maxElevChangeWin)
    maxElevChangeWin = 20;
end
if isempty(maxElevChangeLine)
    maxElevChangeLine = 250;
end
if isempty(maxGeoid)
    maxGeoid = inf;
end
if isempty(minGeoid)
    minGeoid = -inf;
end
if isempty(gridOpt)
    gridOpt = 'nearest';
end


%% Persistent lastFilePath variable sotes recently used path for use as starting path
persistent lastLidarPath lastGPSPath
if isempty(lastLidarPath)
    lastLidarPath = pwd;
elseif lastLidarPath == 0
    lastLidarPath = pwd;
end
if isempty(lastGPSPath)
    lastGPSPath = pwd;
elseif lastGPSPath == 0
    lastGPSPath = pwd;
end

minSamplesPerFilterPeriod = 20;

if isempty(inputDir) || ~exist(inputDir,'dir')
    if usejava('desktop')
        inputDir = uigetdir(lastLidarPath,'Select a dir to convert/extract');
    else
        error('lasSurfElev:MissingArgs','Requires path containing las files as an argument')
    end
end
lastLidarPath = inputDir;

if isempty(outputDir)
    outputDir = inputDir;
end


%% Check up on gps and las files
lasfList = dir([inputDir,filesep,'*.las']);
numfiles = numel(lasfList);

if numfiles>0
    fprintf('lasSurfElev: Found %d las files\n Processing...\n',numfiles);
    
    for ll = 1:numfiles
        fileFields = strsplit(lasfList(ll).name,{'_','.'});
        lasfList(ll).recordTime  = datenum([fileFields{3}, fileFields{4}],'yyyymmddHHMMSS');
        lasfList(ll).recordNum = str2double(fileFields{7}(7:end));
        if isnan(lasfList(ll).recordNum)
            warning('off','backtrace');
            warning('lasSurfElev:InvalidRecordNumber',...
                'Invalid Record number for file %s\n,Check the file name format and fields',lasfList(ll).name);
            warning('on','backtrace');
        end
    end
    
    %sort the files by record number
    [~,sx] = sort([lasfList.recordNum]);
    lasfList = lasfList(sx);
    
    %lets find any gaps in data
    fileperiod = mode(diff([lasfList.recordTime]*86400));
    datafileGaps = [false diff([lasfList.recordTime]*86400)>fileperiod*1.5];
    
    yearkey = {'AN01','20142015';'AN02','20152016';'AN03','20162017';'AN04','20172018'};
    fileFields = strsplit(lasfList(1).name,{'_','.'});
    campaign = fileFields{1};
    flightnumber = fileFields{2};
    campaignyear = yearkey{ismember(yearkey(:,1),campaign),2};
    [~,firstfilenameonly,~] = fileparts(lasfList(1).name);
else
    error('lasSurfElev:noLasFiles','No las files found in the input directory %s',inputDir)
end


if isempty(gpsFile) || ~exist(gpsFile,'file')
    [gpsFile,gpspath] = uigetfile([lastGPSPath,filesep,'*.mat'],'Select the gps .mat file for the flight');
    gpsFile = fullfile(gpspath,gpsFile);
elseif strcmpi(gpsFile,'hestia')
    gpsfList = dir(['/local/data/hestia/trunk/icepod/antarctica/',campaignyear,filesep,...
        'pnt/pod-ins/trajectory/',flightnumber,filesep,'mat/*FINAL*.mat']);
    gpspath = ['/local/data/hestia/trunk/icepod/antarctica/',campaignyear,filesep,...
        'pnt/pod-ins/trajectory/',filesep,flightnumber,filesep,'mat/'];
    if numel(gpsfList) == 1
        gpsFile = [gpspath,gpsfList.name];
    else
        error('lasSurfElev:cannotFindGPSFile','Cannot identify GPS trajectory file!');
    end
end
[lastGPSPath,~,~] = fileparts(gpsFile);


tic
% load
load(gpsFile,'GPSUTC','Lat','Lon','Hell');

% pre-allocate
SurfHell = ones(size(GPSUTC))*NaN;
Reflectance = ones(size(GPSUTC))*NaN;

outMatFileName = [outputDir,filesep,'mat',filesep,firstfilenameonly,'.mat'];
% Got good gps and las files let's make the output dir
if ~exist([outputDir,filesep,'mat'],'dir')
    mkdir([outputDir,filesep,'mat']);
elseif exist(outMatFileName,'file')
    if overwriteOutput
        warning('off','backtrace');
        warning('lasSurfElev:outFileOverwrite','Found existing processed data, they will be overwritten\nPress Ctrl + c to cancel')
        warning('on','backtrace');
        x = warning('query','lasSurfElev:outFileOverwrite');
        if strcmpi(x.state,'on')
            fprintf('lasSurfElev: Waiting %d seconds',5);
            for ss = 5:-1:0
                fprintf('\b\b\b\b\b\b\b\b\b%d seconds',ss);
                pause(1);
            end
            fprintf('\n');
        end
        warning('off','lasSurfElev:outFileOverwrite');
    else
        warning('off','backtrace');
        warning('lasSurfElev:outFileExists','Found existing processed data, processing skipped!')
        warning('on','backtrace');
        return;
    end
end
if ~exist([outputDir,filesep,'txt'],'dir')
    mkdir([outputDir,filesep,'txt']);
end


lastGoodElev = NaN;
%% Go through and read all lass files, filter then grid to get surfElev

for ff = 1:numel(lasfList)
    fprintf('lasSurfElev: File %d of %d \n',ff,numel(lasfList))
    lasfile = [inputDir,filesep,lasfList(ff).name];
    [~,file,~] = fileparts(lasfile);
    lasmatfile = [fullfile(outputDir,file),'.mat'];
    
    % Read laser points x, y and z and some other headers
    lasobj = lasdata(lasfile);
    % Read gps time and intensity
    get_gps_time(lasobj);
    get_intensity(lasobj);
    get_scan_angle(lasobj);
    % convert las file adjusted GPS time to
    pointUtc = adjStdGPStoUTC(double(lasobj.gps_time));
    % las file GPS times seems to be pre-adjusted for leap seconds so add
    % that back in
    pointUtc = pointUtc + getGPSLeapSecs(min(pointUtc));
    
    % read data into some variable we can use and save
    pointHell = lasobj.z;
    pointLon = lasobj.x;
    pointLat = lasobj.y;
    pointReflectance = lasobj.intensity;
    pointAngle = lasobj.scan_angle;
    
    % It appears that the point cloud in las file is not always linear in time
    % sort the point cloud by time
    [b,i] = sort(pointUtc);
    pointHell = pointHell(i);
    pointUtc = b;clear b;
    pointLon = pointLon(i);
    pointLat = pointLat(i);
    pointReflectance = pointReflectance(i);
    pointAngle = pointAngle(i);
    
    if plotPointcloud
        figure(1);
        hold on;
        plot(unix2matlabtime(pointUtc(1:10:end)),pointHell(1:10:end),'.');
    end
    
    numWindows = floor(range(pointUtc)/filterWindow);
    numSamplePerWin = floor(numel(pointHell)/numWindows);
    
    if numSamplePerWin < minSamplesPerFilterPeriod
        warning('off','backtrace');
        warning('lasSurfElev:tooFewSamples',...
            '%d samples available per second; total %d samples in the file %s - Skipped!',numSamplePerWin,numel(pointHell),lasfList(ff).name);
        warning('on','backtrace');
        lastGoodElev = NaN;
        continue;
    end
    
    if datafileGaps(ff)
        lastGoodElev = NaN;
    end
    
    % if remaining data are more than 50% of the numSamplesPerWin,
    % add another window
    if (numel(pointHell) - numWindows*numSamplePerWin)/numSamplePerWin > 0.5
        numWindows = numWindows + 1;
    end
    
    % pre-allocate
    pointStd = zeros(1,numWindows);
    pointMean = zeros(1,numWindows);
    minsurfHell = zeros(1,numWindows);
    maxsurfHell = zeros(1,numWindows);
    
    % Most of the noise (or unwanted values) in elevation are much higher or
    % lower than the 'real' elevation. This noise changes the mean and std of
    % the data significantly but the 'real' elevation points are concentrated
    % within ~2 m area (for +/- 5 deg swath) where as noise is spread across
    % 10s and even 100s of meters. This filtering scheme uses that fact in
    % following steps:
    % 1. Break up data into filterWin second segments (1 by default)
    % 2. If standard deviation for each segment is smaller than thresholdDeviation,
    %    assume clean data, remove any points out side of 6 sigma, skip further
    %    filtering
    % 3. For noisy data, remove any points outside of 3 std
    % 4. For remaining data, get histogram counts and use the value with
    %    maximum count as the true mean surface elevation for the segment
    % 5. Remove any points outside of maxElevChange from the mean surface
    %    elevation
    % 6. Pass data (from 2 or 5) through median spike-supression filter
    % 7. Remove any points outside of 6 sigma
    
    % Surface elevation should be relatively clean at this point, if not Just
    % toss them in the garbage -- just kidding. Keep playing with
    % filterWindow, thresholdDeviation and maxElevChange as needed to get good
    % results
    
    for ffi = 1:numWindows
        
        if ffi == numWindows
            winInds = (ffi-1)*numSamplePerWin+1:numel(pointHell);
        else
            winInds = (ffi-1)*numSamplePerWin+1:min(ffi*numSamplePerWin,numel(pointHell));
        end
        
        % Remove elevation outside of max and min geoid heights
        pointHell(pointHell>maxGeoid) = NaN;
        pointHell(pointHell<minGeoid) = NaN;
        
        % Remove elevation outside of maximum permissible change in the line
        if ~isnan(lastGoodElev)
            pointHell(pointHell>lastGoodElev + maxElevChangeLine | pointHell < lastGoodElev - maxElevChangeLine) = NaN;
        end
        
        pointStd(ffi) = nanstd(pointHell(winInds));
        pointMean(ffi) = nanmean(pointHell(winInds));
        
        if pointStd(ffi) > thresholdDeviation
            badinds = find(pointHell(winInds) > (pointMean(ffi) + 3*pointStd(ffi)) | pointHell(winInds) < (pointMean(ffi) - 3*pointStd(ffi)));
            pointHell(winInds(1) + badinds -1) = NaN;
            [p,values] = histcounts(pointHell(winInds));
            [~,maxInd] = max(p);
            meanVal = (values(maxInd) + values(maxInd + 1))/2;
            minsurfHell(ffi) = meanVal - maxElevChangeWin;
            maxsurfHell(ffi) = meanVal + maxElevChangeWin;
            badinds = find(pointHell(winInds)<minsurfHell(ffi) | pointHell(winInds)>maxsurfHell(ffi));
            pointHell(winInds(1) + badinds -1) = NaN;
        else
            badinds = find(pointHell(winInds) > (pointMean(ffi) + 6*pointStd(ffi)) | pointHell(winInds) < (pointMean(ffi) - 6*pointStd(ffi)));
            pointHell(winInds(1) + badinds -1) = NaN;
        end
        thisGoodElev = nanmean(pointHell(winInds));
        if ~isnan(thisGoodElev)
            lastGoodElev = thisGoodElev;
        end
        
    end
    pointHell = medfilt1(pointHell);
    allStd = nanstd(pointHell);
    allMean = nanmean(pointHell);
    
    badinds = (pointHell > (allMean + 6*allStd) | pointHell < (allMean - 6*allStd));
    pointHell(badinds) = NaN;
    
    % remove NaN values from all data
    nanInds = isnan(pointLat) | isnan(pointLon) | isnan(pointHell);
    pointLat(nanInds) = [];
    pointLon(nanInds) = [];
    pointHell(nanInds) = [];
    pointUtc(nanInds) = []; 
    pointReflectance(nanInds) = [];
    if ~isempty(pointUtc)
        inRange = GPSUTC >= min(pointUtc) & GPSUTC <= max(pointUtc);
    else
        warning('off','backtrace');
        warning('lasSurfElev:NoElevationData','No valid elevation data in %s',lasfile);
        warning('on','backtrace');
        continue
    end
    %% Grid the clean point-cloud and get elevation values for GPS flight track
    SurfHell(inRange) = griddata(pointLon,pointLat,pointHell,Lon(inRange),Lat(inRange),gridOpt);
    Reflectance(inRange) = griddata(pointLon,pointLat,double(pointReflectance),Lon(inRange),Lat(inRange),gridOpt);
    % save point cloud if asked for
    if(savePointcloud)
        save(lasmatfile,'pointHell','pointUtc','pointLon','pointLat','pointAngle','pointReflectance');
    end
end

if plotPointcloud
    figure(1);
    grid;
    datetick2
    xlabel('UTC Time of Day')
    ylabel('Ellipsoidal Height, m')
    filefields = strsplit(lasfList(1).name,'_');
    campaign = filefields{1};
    flight = filefields{2};
    flightdate = filefields{3};
    title([campaign,' ',flight ' Ground Elevation from Lidar']);
    print([outputDir,filesep,campaign,'_',flight,'_',flightdate,'_LidarGroundElev'],'-dpng')
end

goodInds = ~isnan(SurfHell);
% save GroundTrack file
SurfHell = SurfHell(goodInds);
Reflectance = Reflectance(goodInds);
Lat = Lat(goodInds);
Lon = Lon(goodInds);
Hell = Hell(goodInds);
GPSUTC = GPSUTC(goodInds);

outTxtFileName = [outputDir,filesep,'txt',filesep,firstfilenameonly,'.txt'];

%% hardcoded elements - may need to pull these out into config file later
Notes = strsplit(lasfList(1).name,{'_','.'});

save(outMatFileName,'Lat','Lon','Hell','GPSUTC','SurfHell','Reflectance','Notes')

txtfid = fopen(outTxtFileName,'w');
fcloser = onCleanup(@()fclose(txtfid));

fprintf(txtfid,'Platform: Icepod\nCampaign: %s\nFlight: %s\nSensor: %s %s\n',...
    Notes{1},Notes{2},Notes{5},Notes{6});
fprintf(txtfid,'UTCDate,UTCTime,UnixUTC,Lat,Lon,Hell,SurfHell,Reflectance\n');
fprintf(txtfid,'mm/dd/yyyy,HH:MM:SS.FFF,sec,deg,deg,m,m,\n');
for ddk = 1:numel(SurfHell)
fprintf(txtfid,'%s,%0.3f,%0.10f,%0.10f,%0.3f,%0.3f,%d\n',datestr(unix2matlabtime(GPSUTC(ddk)),'mm/dd/yyyy,HH:MM:SS.FFF'),GPSUTC(ddk),...
    Lat(ddk),Lon(ddk),Hell(ddk),SurfHell(ddk),Reflectance(ddk));
end
clear('fcloser');
toc
fprintf('lasSurfElev: Done!\n')
end
